/*
 * mkstorage.h --
 *
 *	This file contains the definition of the Metakit driver for
 *	e4Graph. This file defines the class e4_MetakitStorageImpl which
 *	is derived from e4_StorageImpl.
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Get the definitions from MetaKit. MetaKit is being used as a dependent
 * DLL, so define q4_KITDLL.
 */

#include "mk4.h"
#include "mk4str.h"

/*
 * These constants define the version number of the MK storage driver:
 *
 * The current format version is 1.4:
 */

#define E4_MKSTORAGE_MAJORVER		1
#define E4_MKSTORAGE_MINORVER		4

/*
 * These layout strings describe the format of storage used by e4Graph. The
 * e4Graph format uses several top-level views:
 *
 * e4GraphMarkers	- contains all markers in this storage.
 *			  (removed as of 1.4)
 * e4GraphNodes		- contains all nodes in this storage.
 * e4GraphVertices	- contains all vertices in this storage.
 * e4GraphNames		- contains all marker and vertex names in this storage.
 * e4GraphDoubles	- contains all 64-bit floating point data.
 * e4GraphStrings	- contains all string data in this storage.
 * e4GraphBinary	- contains all binary data in this storage.
 * e4GraphUnused	- contains indices of "unused slot" lists for
 *			  all other views in this storage, as well as
 *			  some metadata and bookkeeping information.
 * e4GraphParents	- contains indices of all parent nodes.
 *
 * Integer data is represented directly in the vertex, in the "row" column.
 */

/*
 * Five versions supported, 1.0, 1.1, 1.2, 1.3 and 1.4:
 */

#define	E4_GRAPHMARKERS1_0 \
	"e4GraphMarkers[nameID:I,next:I,root:I,used:I,userData:I]"
#define	E4_GRAPHNODES1_0 \
	"e4GraphNodes[firstVertex:I,lastVertex:I,vertexCount:I,parentID:I," \
		      "used:I,next:I,refCount:I,nodeMarkers:I,userData:I]"
#define E4_GRAPHVERTICES1_0 \
	"e4GraphVertices[nameID:I,nodeID:I,vertexType:I,row:I,next:I," \
			 "prev:I,used:I,userData:I]"
#define E4_GRAPHDOUBLES1_0 "e4GraphDoubles[d:D,next:I,used:I]"
#define E4_GRAPHSTRINGS1_0 "e4GraphStrings[s:S,next:I,used:I]"
#define E4_GRAPHNAMES1_0 "e4GraphNames[n:S,next:I,used:I]"
#define E4_GRAPHBINARY1_0 "e4GraphBinary[b:B,next:I,used:I]"
#define E4_GRAPHUNUSED1_0 "e4GraphUnused[first:I]"
#define E4_GRAPHPARENTS1_0 "e4GraphParents[nodeID:I,count:I,next:I,used:I]"

#define	E4_GRAPHMARKERS1_1 \
	"e4GraphMarkers[nameID:I,next:I,root:I,flags:I,userData:I]"
#define	E4_GRAPHNODES1_1 \
	"e4GraphNodes[firstVertex:I,lastVertex:I,vertexCount:I,parentID:I," \
		      "flags:I,next:I,refCount:I,nodeMarkers:I,userData:I]"
#define E4_GRAPHVERTICES1_1 \
	"e4GraphVertices[nameID:I,nodeID:I,vertexType:I,row:I,next:I," \
			 "prev:I,flags:I,userData:I]"
#define E4_GRAPHDOUBLES1_1 "e4GraphDoubles[d:D,next:I,flags:I]"
#define E4_GRAPHSTRINGS1_1 "e4GraphStrings[s:S,next:I,flags:I]"
#define E4_GRAPHNAMES1_1 "e4GraphNames[n:S,next:I,flags:I]"
#define E4_GRAPHBINARY1_1 "e4GraphBinary[b:B,next:I,flags:I]"
#define E4_GRAPHUNUSED1_1 "e4GraphUnused[first:I]"
#define E4_GRAPHPARENTS1_1 "e4GraphParents[nodeID:I,count:I,next:I,flags:I]"

#define	E4_GRAPHMARKERS1_2 \
	"e4GraphMarkers[nameID:I,next:I,root:I,flags:I,userData:I]"
#define	E4_GRAPHNODES1_2 \
	"e4GraphNodes[firstVertex:I,lastVertex:I,vertexCount:I,parentID:I," \
		      "flags:I,next:I,refCount:I,nodeMarkers:I,userData:I]"
#define E4_GRAPHVERTICES1_2 \
	"e4GraphVertices[nameID:I,nodeID:I,vertexType:I,row:I,next:I," \
			 "prev:I,nextinparent:I,flags:I,userData:I]"
#define E4_GRAPHDOUBLES1_2 "e4GraphDoubles[d:D,next:I,flags:I]"
#define E4_GRAPHSTRINGS1_2 "e4GraphStrings[s:S,next:I,flags:I]"
#define E4_GRAPHNAMES1_2 "e4GraphNames[n:S,next:I,flags:I]"
#define E4_GRAPHBINARY1_2 "e4GraphBinary[b:B,next:I,flags:I]"
#define E4_GRAPHUNUSED1_2 "e4GraphUnused[first:I]"
#define E4_GRAPHPARENTS1_2 "e4GraphParents[nodeID:I,count:I,next:I,flags:I," \
					   "vertexchain:I]"

#define	E4_GRAPHMARKERS1_3 \
	"e4GraphMarkers[nameID:I,next:I,root:I,flags:I,userData:I]"
#define	E4_GRAPHNODES1_3 \
	"e4GraphNodes[firstVertex:I,lastVertex:I,vertexCount:I,parentID:I," \
		      "flags:I,next:I,refCount:I,nodeMarkers:I,userData:I," \
		      "detachedvertices:I]"
#define E4_GRAPHVERTICES1_3 \
	"e4GraphVertices[nameID:I,nodeID:I,vertexType:I,row:I,next:I," \
			 "prev:I,nextinparent:I,flags:I,userData:I]"
#define E4_GRAPHDOUBLES1_3 "e4GraphDoubles[d:D,next:I,flags:I]"
#define E4_GRAPHSTRINGS1_3 "e4GraphStrings[s:S,next:I,flags:I]"
#define E4_GRAPHNAMES1_3 "e4GraphNames[n:S,next:I,flags:I]"
#define E4_GRAPHBINARY1_3 "e4GraphBinary[b:B,next:I,flags:I]"
#define E4_GRAPHUNUSED1_3 "e4GraphUnused[first:I]"
#define E4_GRAPHPARENTS1_3 "e4GraphParents[nodeID:I,count:I,next:I,flags:I," \
					   "vertexchain:I]"

/*
 * As of version 1.4 there is no e4GraphMarkers view:
 */

#define	E4_GRAPHNODES1_4 \
	"e4GraphNodes[firstVertex:I,lastVertex:I,vertexCount:I,parentID:I," \
		      "flags:I,next:I,refCount:I,userData:I," \
		      "detachedvertices:I]"
#define E4_GRAPHVERTICES1_4 \
	"e4GraphVertices[nameID:I,nodeID:I,vertexType:I,row:I,next:I," \
			 "prev:I,nextinparent:I,flags:I,userData:I]"
#define E4_GRAPHDOUBLES1_4 "e4GraphDoubles[d:D,next:I,flags:I]"
#define E4_GRAPHSTRINGS1_4 "e4GraphStrings[s:S,next:I,flags:I]"
#define E4_GRAPHNAMES1_4 "e4GraphNames[n:S,next:I,flags:I]"
#define E4_GRAPHBINARY1_4 "e4GraphBinary[b:B,next:I,flags:I]"
#define E4_GRAPHUNUSED1_4 "e4GraphUnused[first:I]"
#define E4_GRAPHPARENTS1_4 "e4GraphParents[nodeID:I,count:I,next:I,flags:I," \
					   "vertexchain:I]"

/*
 * These constants define indices into the e4GraphUnused view, used for storing
 * the unused-list indices for each kind of unused slot list in a storage.
 */

#define E4_GRAPHSTORAGEMAJORVER		0
#define E4_GRAPHSTORAGEMINORVER		1
#define	E4_GRAPHFIRSTUNUSEDMARKER	2
#define E4_GRAPHFIRSTUNUSEDNODE		3
#define E4_GRAPHFIRSTUNUSEDVERTEX	4
#define E4_GRAPHFIRSTUNUSEDDOUBLE	5
#define E4_GRAPHFIRSTUNUSEDSTRING	6
#define E4_GRAPHFIRSTUNUSEDBINARY	7
#define E4_GRAPHFIRSTUNUSEDNAME		8
#define E4_GRAPHFIRSTUNUSEDPARENT	9
#define E4_GRAPHHASHCODE		10
#define E4_GRAPHUNREACHABLENODES	11
#define E4_GRAPHDEFERGC			12
#define E4_GRAPHROOTNODE		13

/*
 * This constant defines the size of the e4GraphUnused view.
 */

#define E4_GRAPHUNUSEDSIZE	20

/*
 * Special value #defines:
 */

#define E4_GRAPHINCREMENT	128	/* By how much to increase the size
					 * of a view if there is no free
                                         * space in it. */

#define E4_VERTICESINCREMENT	128	/* By how much to increase the size of
					 * the vertices view if there is no
                                         * free space in it. */

/*
 * The e4_MetakitStorageImpl class:
 */

class e4_MetakitStorageImpl: public e4_StorageImpl 
{
private:
    c4_Storage *storage;		/* The actual MetaKit storage. */
    c4_View parents;			/* Cached view of the parents
					 * of all nodes in this storage. */
    c4_View markers;			/* Cached view of the markers
                                         * in this storage.
					 *
					 * NOTE: We keep this view because
					 * we may need to convert storages
					 * from an older format to 1.4 or
					 * later. */    
    c4_View nodes;			/* Cached view of the nodes
                                         * in this storage. */
    c4_View vertices;			/* Cached view of the vertices
                                         * in this storage. */
    c4_View doubles;			/* This view contains the
                                         * values for all double-valued
                                         * vertices in this storage. */
    c4_View strings;			/* This view contains the
                                         * values of all string
                                         * vertices in this storage. */
    c4_View names;			/* This view contains all
                                         * the strings used as
                                         * vertex names in this storage. */
    c4_View binary;			/* This view contains the
                                         * values of all binary
                                         * vertices in this storage. */
    c4_View unused;			/* This view contains the values of
                                         * all unused-list indices for
                                         * this storage. */

    int statistics[(int) E4_SPLAST][(int) E4_SSLAST];
					/* Space for collecting statistics. */

    /*
     * Storage used for deletion mechanism:
     */

    e4_IntStack *idStack1, *idStack2;

    /*
     * Is GC needed?
     */

    bool needsGC;

private:
    e4_MetakitStorageImpl();		/* The default constructor is private
					 * to prevent it from being called. */
    e4_MetakitStorageImpl(const e4_MetakitStorageImpl &);
					/* Copy constructor cannot be
					 * called either. */
    void operator= (const e4_MetakitStorageImpl *);
					/* Nor assignment. */


    /*
     * GC mechanism:
     */

    void InitGC();
    void SpanReachableNodes();
    void SeedReachableNodesFromReferencedNodes();
    void SeedReachableNodesFromReferencedVertices();
    void SpanSeededNodes();

    void CollectUnreachableEntities();
    void CollectUnreachableNodes();
    void CollectUnreachableVertices();

    /*
     * Fire detach events for newly detached entites, after a GC.
     */

    void FireEventsForNewlyDetached();
    void FireEventsForNewlyDetachedNodes();
    void FireEventsForNewlyDetachedVertices();

    /*
     * These operations are only for use by the implementation of
     * e4_MetakitStorageImpl.
     */

    void MakeParentSpace();
    void MakeNodeSpace();
    void MakeVertexSpace();
    void MakeDoubleSpace();
    void MakeStringSpace();
    void MakeBinarySpace();
    void MakeNameSpace();

    /*
     * These operations free up a previously used item given the unique
     * identifier for the item. The operations return true if the item
     * was successfully freed.
     */

    bool FreeDouble(int index);
    bool FreeString(int index);
    bool FreeBinary(int index);

    /*
     * These methods are helpers that implement different methods to visit
     * vertices in a storage:
     */

    e4_VertexImpl *FindNextVertexStorage(int vertexID, int vf, int nameID,
					 e4_VertexType ft) const;
    e4_VertexImpl *FindNextVertexNode(int VertexID, int vf, int nameID,
				      e4_VertexType ft, int nodeID) const;
    e4_VertexImpl *FindNextVertexNodeRandom(int vertexID, int vf, int nameID,
					    e4_VertexType ft, int nodeID)
	const;

    e4_VertexImpl *FindNextVertexStorageNone(int vertexID) const;
    e4_VertexImpl *FindNextVertexStorageName(int vertexID, int nameID) const;
    e4_VertexImpl *FindNextVertexStorageType(int vertexID, e4_VertexType ft) 
	const;
    e4_VertexImpl *FindNextVertexStorageBoth(int vertexID, int nameID,
					     e4_VertexType ft) const;

    e4_VertexImpl *FindNextVertexNodeNone(int vertexID, int nodeID)const;
    e4_VertexImpl *FindNextVertexNodeName(int vertexID, int nameID,
					  int nodeID) const;
    e4_VertexImpl *FindNextVertexNodeType(int vertexID, e4_VertexType ft,
					  int nodeID) const;
    e4_VertexImpl *FindNextVertexNodeBoth(int vertexID, int nameID,
					  e4_VertexType ft, int nodeID) const;

    e4_VertexImpl *FindNextVertexNRNone(int vertexID, int nodeID) const;
    e4_VertexImpl *FindNextVertexNRName(int vertexID, int nameID,
					int nodeID) const;
    e4_VertexImpl *FindNextVertexNRType(int vertexID, e4_VertexType ft,
					int nodeID) const;
    e4_VertexImpl *FindNextVertexNRBoth(int vertexID, int nameID,
					e4_VertexType ft, int nodeID) const;

    int FindNextVertexIndexInNode(int vertexID, int nodeID) const;
    
    /*
     * Populate the name hash table with the known names.
     */

    void PopulateNameHash();
    
    /*
     * Private constructor used by GetStorage.
     */

    e4_MetakitStorageImpl(const char *fname);
    
    /*
     * Helper functions for moving vertices around.
     */

    int SpliceOut(int vertexID, bool updateChild);
    void SpliceIn(int vertexID, int newParent, int oldParent, int rank);

    /*
     * Compute the total number of slots in a specific space in the storage.
     */

    void UpdateAvailable(e4_Space sp);

    /*
     * Compute the number of slots used in a specific space in the storage.
     */

    void UpdateUsed(e4_Space sp);

    /*
     * Initialize the storage.
     */

    bool Initialize();

    /*
     * Free the value of a vertex.
     */

    void FreeVertexValue(int vertexID);

    /*
     * Remove a parent from a node.
     */

    void RemoveParent(int childID, int parentID, int vertexID,
		      bool GCOrDetach);

    /*
     * Update the format from 1.0 to 1.1:
     */

    void UpdateFormat1_0to1_1();

    /*
     * Update the format from 1.1 to 1.2:
     */

    void UpdateFormat1_1to1_2();

    /*
     * Update the format from 1.2 to 1.3:
     */

    void UpdateFormat1_2to1_3();

    /*
     * Update the format from 1.3 to 1.4:
     */

    void UpdateFormat1_3to1_4();

    /*
     * Clean up any remaining detached entities after the last commit.
     */

    void CleanupDetached();

    /*
     * Internal, more efficient way to detach a vertex.
     */

    void DetachVertexInternal(int vertexID, int childID);

public:

    /*
     * The destructor:
     */

    virtual ~e4_MetakitStorageImpl();
    
    /*
     * This method is called to create a new instance of
     * e4_MetakitStorageImpl.
     */

    static e4_StorageImpl *GetStorage(const char *fname);
    
    /*
     * These public methods are declared virtual pure in e4_StorageImpl and
     * must be implemented by subclasses.
     */
    
    virtual bool Commit();
    virtual bool IsValid() const;
    virtual bool SetRootNodeID(int nodeID);
    virtual int  GetRootNodeID();
    virtual bool IsRootNodeID(int nodeID);

    /*
     * Get a value for a statistic being collected for this storage.
     */

    virtual bool GetStatistic(e4_Space sp, e4_SpaceStat st, int &v);

    /*
     * Return the hash code for this storage.
     */

    virtual int HashCode() const;

    /*
     * Is GC deferred?
     */

    virtual bool IsGCDeferred() const;
    virtual bool SetGCDeferral(bool newval);

    /*
     * Do a GC.
     */

    virtual void DoGC();

    /*
     * Is a GC needed?
     */

    virtual bool NeedsGC();

    /*
     * Set the flag indicating whether GC is needed.
     */

    void SetNeedsGC(bool needs);

protected:

    /*
     * These protected methods are declared virtual pure in e4_StorageImpl
     * and must be implemented by subclasses.
     */

    /*
     * Count how many vertices there are in a node, given the nodeID.
     */

    virtual int VertexCountFromNodeID(int nodeID) const;

    /*
     * Count how many vertices in a node with this nodeID have the given
     * name, type or value.
     */

    virtual int VertexCountWithNameFromNodeID(int nodeID, const char *nm)
	const;
    virtual int VertexCountWithTypeFromNodeID(int nodeID, e4_VertexType tp)
	const;
    virtual int VertexCountWithValueFromNodeID(int nodeID, e4_Value &v)
	const;

    /*
     * Find the rank of a child node in the vertex list of a node.
     */

    virtual int GetRankOfChildNode(int parentID, int childID) const;
    
    /*
     * Set a vertex whose vertex ID is given.
     */
    
    virtual bool SetVertex(int vertexID, int nameID, int vertexType,
			   int itemID);

    /*
     * Add a vertex according to the rank and io arguments to a node whose
     * node ID is given.
     */
    
    virtual int AddVertex(int nodeID, e4_InsertOrder io, int &rank);
    
    /*
     * Find the rank of a vertex with a given index in a node whose node ID
     * is given.
     */

    int RankFromVertexID(int nodeID, int vertexID) const;

    /*
     * Find the nth vertex with a given name in a list of vertices.
     */
    
    virtual int VertexIDFromNthVertex(int nodeID, int nameID,
				      int nth, int &rank) const;

    /*
     * Get the next vertex after one identified by vertexID in
     * the node containing this vertex.
     */

    virtual e4_VertexImpl *NextVertexAfter(int num, int vertexID) const;

    /*
     * Get the previous vertex before one identified by vertexID in
     * the node containing this vertex.
     */

    virtual e4_VertexImpl *PrevVertexBefore(int num, int vertexID) const;

    /*
     * Given the rank of a vertex in this node, find the corresponding
     * vertex ID.
     */

    virtual int VertexIDFromRank(int nodeID, int rank) const;

    /*
     * Get the parent node of a node or the parent ID, given the child node ID.
     */

    virtual e4_NodeImpl *GetParentNode(int childID, int nth) const;
    virtual int GetParentNodeID(int childID, int nth) const;

    /*
     * Get the vertex ID for the vertex containing this node in its parent.
     */

    virtual int GetVertexIDInParent(int parentID, int nodeID) const;
    
    /*
     * Get vertex IDs for the first or last vertex in a node, given 
     * the node ID.
     */

    virtual int GetFirstVertexID(int nodeID) const;
    virtual int GetLastVertexID(int nodeID) const;

    /*
     * Get the e4_VertexImpl instance representing the vertex following a
     * vertex identified by ID in the node containing that vertex.
     */

    virtual e4_VertexImpl *FindNextVertex(int vertexID, e4_VisitMethod vm,
					  int vf, int nameID, int nodeID,
					  e4_VertexType typeID)
	const;

    /*
     * Get the e4_NodeImpl instance representing the node following a
     * node identified by ID.
     */

    virtual e4_NodeImpl *FindNextNode(int nodeID) const;

    /*
     * Given a vertex ID, find the e4_NodeImpl instance representing the node
     * that contains that vertex.
     */

    virtual e4_NodeImpl *ContainingNodeFromVertexID(int vertexID) const;

    /*
     * Is the node identified by nodeID valid?
     */

    virtual bool IsLegalNodeID(int nodeID) const;

    /*
     * Check that a vertex ID is legal.
     */

    virtual bool IsLegalVertexID(int index) const;

    /*
     * Get the value of a vertex whose vertex ID is given.
     */

    virtual bool GetVertexByIndex(int index, e4_ValueImpl *&v) const;
    virtual bool GetVertexByIndex(int index, e4_NodeImpl *&n) const;
    virtual bool GetVertexByIndex(int index, int &v) const;
    virtual bool GetVertexByIndex(int index, double &f) const;
    virtual bool GetVertexByIndex(int index, const char *&s) const;
    virtual bool GetVertexByIndex(int index, const void *&bytes,
				  int &nbytes) const;

    /*
     * Set the value of a vertex whose vertex ID is given.
     */

    virtual bool SetVertexByIndex(int index, int v);
    virtual bool SetVertexByIndex(int index, double f);
    virtual bool SetVertexByIndex(int index, const char *s);
    virtual bool SetVertexByIndex(int index, const void *bytes, 
				  int nbytes);
    virtual bool SetVertexByIndexToNode(int index, int childID);

    /*
     * Rename a vertex whose index is given.
     */

    virtual bool RenameVertexByVertexID(int vertexID, const char *newname);

    /*
     * Return the vertex type and name of the vertex whose ID is given.
     */

    virtual e4_VertexType VertexTypeFromVertexID(int vertexID) const;
    virtual int NodeIDFromVertexID(int vertexID) const;
    virtual const char *VertexNameFromVertexID(int vertexID) const;
    virtual int NameIDFromVertexID(int vertexID) const;

    /*
     * Set and get the user data associated with an entity.
     */

    virtual bool SetNodeUserData(int nodeID, int userData) const;
    virtual bool SetVertexUserData(int vertexID, int userData) const;
    virtual bool GetNodeUserData(int nodeID, int &userData) const;
    virtual bool GetVertexUserData(int vertexID, int &userData) const;

    /*
     * Retrieve the value of an entry in one of the type-specific
     * views. The value is returned in an output parameter. If the
     * row contains an in-use value, true is returned and the output
     * parameter is set. Otherwise false is returned and the
     * output parameter is not modified.
     */

    virtual bool GetNode(int index, e4_NodeImpl *&value) const;
    virtual bool GetInt(int index, int &value) const;
    virtual bool GetDouble(int index, double &value) const;
    virtual bool GetString(int index, char *&value) const;
    virtual bool GetBinary(int index, void *&value, int &nbytes) const;
    virtual bool GetVertexName(int index, const char *&value) const;

    /*
     * Specialized version of GetNode that also returns an indication
     * whether the node is newly created.
     */

    virtual bool GetNode(int index, e4_NodeImpl *&value, bool *isNew) const;

    /*
     * Detach the node or vertex identified by the given ID.
     */

    virtual bool DetachNodeByID(int id);
    virtual bool DetachVertexByID(int id);

    /*
     * Detach the first vertex in parentID that has childID as its node value.
     */

    virtual bool DetachFirstVertexWithNode(int parentID, int childID);

    /*
     * Is the node or vertex identified by the given ID detached?
     */

    virtual bool IsDetachedNodeID(int id);
    virtual bool IsDetachedVertexID(int id);

    /*
     * Add items to each specific free list.
     */

    virtual void AddToUnusedNodesList(int i);
    virtual void AddToUnusedVerticesList(int i);
    virtual void AddToUnusedIntsList(int i);
    virtual void AddToUnusedDoublesList(int i);
    virtual void AddToUnusedStringsList(int i);
    virtual void AddToUnusedNamesList(int i);
    virtual void AddToUnusedBinaryList(int i);
    virtual void AddToUnusedParentList(int i);

    /*
     * Add a new entry to one of the type-specific views. The operation
     * returns the index of the new row in the view.
     */

    virtual int ReserveNodeID();
    virtual int ReserveVertexID(const char *nm);
    virtual int AddInt(int value);
    virtual int AddDouble(double value);
    virtual int AddString(const char *value);
    virtual int AddBinary(const void *value, int nbytes);
    virtual int AddName(const char *nm);

    /*
     * Add the node identified by "parentID" to the list of nodes that
     * contain the node identified by "childID".
     */

    virtual void AddParent(int childID, int parentID, int vertexID);

    /*
     * Move the vertex identified by vertexID into the node identified by
     * nodeID at the given rank.
     */

    virtual bool MoveVertex(int vertexID, int nodeID, int rank);

    /*
     * Copy the e4Graph related data from this storage to another.
     */

    virtual bool CopyTo(e4_StorageImpl *impl);

    /*
     * Destroy the physical storage for this storage and make it invalid.
     */

    virtual void Destroy();

    /*
     * How many parents does the given node have?
     */

    virtual int ParentCount(int nodeID) const;

    /*
     * How many times does the node with the given ID appear as the value
     * of a vertex in this storage?
     */

    virtual int OccurrenceCount(int nodeID) const;

    /*
     * How many times does the node with the given child ID appear as
     * the value of a vertex in the node with the given parent ID?
     */

    virtual int OccurrenceCount(int nodeID, int parentID) const;

    /*
     * What is the rank of the node with the given parent ID for the
     * node with the given child ID?
     */

    virtual int ParentRank(int nodeID, int parentID) const;
};
